import logging
import os
import time
import datetime

from nubium_utils.custom_exceptions import NoMessageError, ConsumeMessageError
from nubium_utils.metrics import MetricsManager


LOGGER = logging.getLogger(__name__)


def _wait_until_message_time(message):
    """
    Wait until the message's timestamp + the deployments offset before handling

    :param message: the kafka message being handled
    :return: None
    """
    message_time = message.timestamp()[1] + (
            int(os.environ['TIMESTAMP_OFFSET_MINUTES']) * 60
    )

    # Divide to get in normal datetime format, since it's milliseconds
    message_time /= 1000

    wait_time = message_time - datetime.datetime.timestamp(
        datetime.datetime.utcnow())

    if wait_time > 0:
        time.sleep(wait_time)


def consume_message(consumer, timeout, monitor: MetricsManager):
    """
    Consumes a message from the broker while handling errors and waiting if necessary

    Polls the broker for a message using the given timeout.
    If there are no messages to consume, either because None is returned
    or the message error is the no messages error,
    raises a NoMessageError.

    If the message is returned with a breaking error,
    raises a ConsumeMessageError.

    If the message is valid, waits until the message's timestamp plus
    the current process's time offset before handling the message.
    This allows retry deployments to wait in a non-blocking fashion

    If the message is valid, then the message is returned
    """

    message = consumer.poll(timeout)

    if message is None:
        LOGGER.info("No messages to consume")
        raise NoMessageError("No messages to consume")

    try:
        guid = [item[1] for item in message.headers() if item[0] == 'guid'][0].decode()
        LOGGER.info(f"Message consumed, GUID: {guid}")
    except AttributeError:
        if "object has no attribute 'headers'" in str(message.error()):
            LOGGER.info("Message consumed. No headers, so no guid is available to log.")
    except TypeError:
        LOGGER.info("Message consumed. Headers are None, so no guid is available to log.")
    except IndexError:
        LOGGER.info("Message consumed. Headers found, but no guid is available to log.")

    # If message is None, it can mean that the poll operation timed out,
    # or that there are no more messages to consume

    if message.error():
        if "Broker: No more messages" in str(message.error()):
            LOGGER.warning("Consumer error: %s", message.error())
            raise NoMessageError(message.error())
        else:
            raise ConsumeMessageError(message.error())

    # Wait until message time if using a retry process
    _wait_until_message_time(message)

    # Increment the metric for consumed messages by one
    monitor.inc_messages_consumed(1, message.topic())

    return message


def commit_message(consumer):
    """
    A convenience method to ensure messages are committed/handled synchronously.
    :param consumer: A confluent-kafka Consumer object
    :type consumer: confluent_kafka.Consumer or confluent_kafka.avro.AvroConsumer
    """
    consumer.commit(asynchronous=False)




class NoMessageError(Exception):
    """
    No message was returned when consuming from the broker
    """
    pass


class NetworkError(Exception):
    """
    Generic network errors
    """
    pass


class ConsumeMessageError(Exception):
    """
    Non-trivial errors when consuming from the broker
    """
    pass


class MessageValueException(Exception):
    """
    Represents a significant error in the value of the message
    """
    pass


class ProduceHeadersException(Exception):
    """
    Represents a significant error in the value of the message
    """
    pass


class ProducerTimeoutFailure(Exception):
    """ Exception signifying a producer failing to communicate with the broker(s). """
    def __init__(self, *args):
        if not args:
            default_message = "Producer could not reach the broker(s) when attempting to send a message."
            args = (default_message,)
        super().__init__(*args)


# Allows for testing Faust functions via pytest (handles async) while stylizing your tests as unittests

from unittest.mock import Mock


class AsyncMock(Mock):
    async def __call__(self, *args, **kwargs):
        return super(AsyncMock, self).__call__(*args, **kwargs)


class AsyncTestCase(object):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def setup_method(self):
        self.setUp()

    def teardown_method(self):
        self.tearDown()

    def __getattribute__(self, item):
        method = object.__getattribute__(self, item)
        if method is None:
            raise AttributeError(f"Unknown object as no attribute '{item}'")
        return method



import sys
import logging
import os
import threading


def _init_handler():
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(_get_loglevel())

    handler.setFormatter(
        logging.Formatter(
            "{process} - {filename:15} - {asctime} - {name} - {levelname} - {message}",
            style="{"
        )
    )
    return handler


def _get_loglevel():
    return os.environ.get('LOGLEVEL', 'CRITICAL')


def init_logger(name='nubium-utils'):
    logger = logging.getLogger(name)
    logger.setLevel(_get_loglevel())
    logger.addHandler(_init_handler())
    logger.propagate = False
    return logger


def init_error_counter_loggers(name):
    logging.setLoggerClass(LogCounter)
    logger = logging.getLogger(name)
    return logger


class LogCounter(logging.Logger):
    """ Overrides logging.Logger to increment a counter when an error or a warning with any of the
    specified strings are logged. """
    def __init__(self, name, level=logging.NOTSET):
        self._count = 0
        self._countLock = threading.Lock()

        return super(LogCounter, self).__init__(name, level)

    @property
    def error_count(self):
        return self._count

    def kill_if_necessary(self):
        if self.name in os.environ['LOGGERS_TO_WATCH'].split(','):
            self._countLock.acquire()
            if self._count >= 1:
                os._exit(1)
            self._countLock.release()

    def warning(self, msg, *args, **kwargs):
        if any(error_string in msg for error_string in os.environ['ERROR_STRINGS'].split(',')):
            self._countLock.acquire()
            self._count += 1
            self._countLock.release()
            self.kill_if_necessary()

        return super(LogCounter, self).warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        if any(error_string in msg for error_string in os.environ['ERROR_STRINGS'].split(',')):
            self._countLock.acquire()
            self._count += 1
            self._countLock.release()
            self.kill_if_necessary()

        return super(LogCounter, self).error(msg, *args, **kwargs)



import unittest
from unittest import mock
import socket

from prometheus_client import CollectorRegistry

from nubium_utils.metrics import MetricsManager, MetricsPusher


class TestMonitor(unittest.TestCase):

    def test_metrics_manager_init(self):
        """
        A metrics_manager initialized with all of it's arguments should succeed
        """
        metrics_manager = MetricsManager(job='test', app='test', registry=CollectorRegistry(), metrics_pusher=mock.MagicMock())
        assert metrics_manager

    def test_registry_auto_create(self):
        """
        When a metrics_manager is created without a registry arg, it initializes one
        """
        metrics_manager = MetricsManager(job='test', app='test', metrics_pusher=mock.MagicMock())
        assert metrics_manager.registry.__class__ == CollectorRegistry

    def test_gauge_creation(self):
        """
        The metrics_manager class should create 3 gauges with specific labels
        """
        metrics_pusher = mock.MagicMock()
        metrics_manager = MetricsManager(job='test', app='test', metrics_pusher=metrics_pusher)

        assert metrics_manager.messages_consumed._labelnames == ('app', 'job', 'topic')
        assert metrics_manager.message_errors._labelnames == ('app', 'job', 'exception')
        assert metrics_manager.messages_produced._labelnames == ('app', 'job', 'topic')

    def test_increment_gauges(self):
        """
        A utility function should increment gauges with specific label values
        """
        metrics_manager = MetricsManager(job='test-job', app='test-app', metrics_pusher=mock.MagicMock())
        metrics_manager.inc_messages_consumed(1, 'test-topic')
        assert metrics_manager.messages_consumed._metrics[('test-app', 'test-job', 'test-topic')]._value._value == 1

        metrics_manager.inc_messages_produced(1, 'test-topic')
        assert metrics_manager.messages_produced._metrics[('test-app', 'test-job', 'test-topic')]._value._value == 1

    def test_increment_message_errors(self):
        """
        Method should increase metric with matching exception label value
        """
        metrics_manager = MetricsManager(job='test-job', app='test-app', metrics_pusher=mock.MagicMock())
        metrics_manager.inc_message_errors(ValueError('test-value-error'))
        assert metrics_manager.message_errors._metrics[('test-app', 'test-job', 'ValueError')]._value._value == 1


class TestMetricsPusher(unittest.TestCase):
    """
    Tests for the class that manages pushing metrics to a metrics cache
    """

    def test_initialization(self):
        """
        Class should initialize with correct arguments
        """
        metrics_pusher = MetricsPusher('test-job', 'test-service-name', 'test-service-port', 'test-pod-port')
        assert metrics_pusher is not None

    def test_set_gateways(self):
        """
        Set gateways returns 2 IP addresses when called with mocked values
        """

        metrics_pusher = MetricsPusher('test-job', 'test-service-name', 'test-service-port', 'test-pod-port')

        # Mock out the socket.getaddrinfo call return value
        addrinfo_mock = [
            (socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_DGRAM, 17, '', ('127.0.0.1', 8080)),
            (socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_DGRAM, 6, '', ('127.0.0.1', 8080)),
            (socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_DGRAM, 17, '', ('127.0.0.2', 8080)),
            (socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_DGRAM, 6, '', ('127.0.0.2', 8080)),
        ]

        with mock.patch('nubium_utils.metrics.metrics_pusher.socket') as socket_patch:
            socket_patch.getaddrinfo.return_value = addrinfo_mock
            metrics_pusher.set_metrics_pod_ips()
            assert metrics_pusher.metrics_pod_ips == {'127.0.0.1:test-pod-port', '127.0.0.2:test-pod-port'}

    def test_push_metrics(self):
        """
        Assert that the push metrics function is called for every gateway
        """

        metrics_pusher = MetricsPusher('test-job', 'test-service-name', 'test-service-port', 'test-pod-port')

        metrics_pusher.metrics_pod_ips = {
            'test-ip-address-1:test-port',
            'test-ip-address-2:test-port'
        }

        with mock.patch('nubium_utils.metrics.metrics_pusher.push_to_gateway') as push_to_gateway_patch:
            registry = CollectorRegistry()
            metrics_pusher.push_metrics(registry)
            assert unittest.mock.call('test-ip-address-1:test-port', job='test-job', registry=registry) in push_to_gateway_patch.call_args_list
            assert unittest.mock.call('test-ip-address-2:test-port', job='test-job', registry=registry) in push_to_gateway_patch.call_args_list



import os

from .general_utils import generate_guid, parse_headers, log_and_raise_error
from .metrics import MetricsPusher, MetricsManager
from .logging_utils import init_logger, init_error_counter_loggers

init_logger(__name__)

if 'DO_ERROR_COUNTING' in os.environ:
    for logger in os.environ['LOGGERS_TO_WATCH'].split(','):
        init_error_counter_loggers(logger)
